# coding=utf-8
from gensim.models import word2vec
from metrics import *
import numpy as np
import time
from collections import defaultdict
from LogisticMatrixFactorization import *

dataset = 'Gowalla'
train_path = './' + dataset + '/' + dataset + '_train.txt'
test_path = './' + dataset + '/' + dataset + '_test.txt'
checkin_path = './' + dataset + '/' + dataset + '_checkins.txt'
size_path = './' + dataset + '/' + dataset + '_data_size.txt'

embedding_dim = 170
win_size = 2
gamma = 0.1


def gen_corpus():
    # return gen_seq(checkin_path, train_dic), gen_seq(test_path)
    return gen_new_train(), gen_seq(test_path)


def gen_new_train():
    train_seq = gen_seq(train_path)
    check_seq = []
    user = []
    seq_dic = {}
    with open(checkin_path) as fp:
        pre_user = 0
        seq = []
        for line in fp.readlines():
            uid, lid, ctime = line.strip('\n').strip('\r').split('\t')
            if lid not in train_seq[int(uid)]:
                continue
            if int(uid) == pre_user:
                seq.append(lid)
            else:
                check_seq.append(seq)
                user.append(str(pre_user))
                seq = [lid]
                pre_user = int(uid)
        check_seq.append(seq)
        user.append(str(pre_user + 1))
    return check_seq


def read_ground_truth():
    ground_truth = defaultdict(set)
    truth_data = open(test_path, 'r').readlines()
    for eachline in truth_data:
        uid, lid, _ = eachline.strip().split()
        uid, lid = int(uid), int(lid)
        ground_truth[uid].add(lid)
    return ground_truth


def read_training_data():
    train_data = open(train_path, 'r').readlines()
    training_tuples = set()
    for eachline in train_data:
        uid, lid, _ = eachline.strip().split()
        uid, lid, = int(uid), int(lid)
        training_tuples.add((uid, lid))
    return training_tuples

def gen_seq(path, train_dic=None):
    check_seq = []
    user = []
    seq_dic = {}
    with open(path) as fp:
        pre_user = 0
        seq = []
        for line in fp.readlines():
            uid, lid, fre = line.strip('\n').strip('\r').split('\t')
            if train_dic:
                if lid not in train_dic.keys():
                    continue
            if int(uid) == pre_user:
                seq.append(lid)
            else:
                check_seq.append(seq)
                user.append(str(pre_user))
                seq = [lid]
                pre_user = int(uid)
        check_seq.append(seq)
        user.append(str(pre_user + 1))
    return check_seq


def get_train_test_dic(train_path, test_path):
    train_dic = {}
    test_dic = {}
    with open(train_path) as fp:
        for line in fp.readlines():
            uid, lid, ctime = line.strip('\n').strip('\r').split('\t')
            train_dic[lid] = 1
    print('train_keys:', len(train_dic.keys()))
    cnt = 0
    with open(test_path) as fp:
        for line in fp.readlines():
            uid, lid, ctime = line.strip('\n').strip('\r').split('\t')
            test_dic[lid] = 1
            if lid not in train_dic.keys():
                # print(lid)
                cnt += 1

    print('test_keys:', len(test_dic.keys()))
    print('cnt:', cnt)
    return train_dic, test_dic


def logistic(A):
    eA = np.exp(A)
    return eA / (1 + eA)

def export_train_seq(train_seq):
    with open('./' + dataset + '/' + dataset + '_corpus.txt', 'w') as fp:
        for seq in train_seq:
            for poi in seq:
                fp.write(str(poi) + ' ')
            fp.write('\n')
    print('export done!')


if __name__ == '__main__':
    st_time = time.time()

    user_num, poi_num = open(size_path, 'r').readlines()[0].strip('\n').split()
    user_num, poi_num = int(user_num), int(poi_num)
    counts, _ = load_matrix(train_path, num_users=user_num, num_items=poi_num)
    LMF = LogisticMF(counts=counts, num_factors=10)
    LMF.train_model()

    training_tuples = read_training_data()
    ground_truth = read_ground_truth()

    print('rec_score time:', time.time() - st_time)

    train_seq, test_seq = gen_corpus()
    export_train_seq(train_seq)

    train_dic, test_dic = get_train_test_dic(train_path, test_path)
    print('len_seq：', len(train_seq))

    result = open(dataset + "_res.txt", 'w')

    all_uids = list(range(user_num))
    all_lids = list(range(poi_num))

    model = word2vec.Word2Vec(train_seq, size=embedding_dim, iter=30, sg=1, hs=1, min_count=1, window=win_size)
    # list for different ks
    precision_5, recall_5 = [], []
    precision_10, recall_10 = [], []
    precision_15, recall_15 = [], []
    precision_20, recall_20 = [], []
    for cnt, uid in enumerate(all_uids):
        if uid in ground_truth:
            actual = ground_truth[uid]

            predicted = model.wv.most_similar(positive=train_seq[uid], topn=len(model.wv.vocab.keys()))
            predicted_dic = dict(predicted)

            overall_scores = [
                logistic(predicted_dic[str(lid)]) + gamma * LMF.predict_logistic(uid, lid)
                if ((uid, lid) not in training_tuples) and str(lid) in predicted_dic.keys() else -1
                for lid in all_lids]
            overall_scores = np.array(overall_scores)
            predicted = list(reversed(overall_scores.argsort()))[:20]

            # calculate the average of different k
            precision_5.append(precisionk(actual, predicted[:5]))
            recall_5.append(recallk(actual, predicted[:5]))

            precision_10.append(precisionk(actual, predicted[:10]))
            recall_10.append(recallk(actual, predicted[:10]))

            precision_15.append(precisionk(actual, predicted[:15]))
            recall_15.append(recallk(actual, predicted[:15]))

            precision_20.append(precisionk(actual, predicted[:20]))
            recall_20.append(recallk(actual, predicted[:20]))

        # print("cnt/total: [{}/{}]".format(cnt, len(all_uids)))
        print("pre@5, rec@5: [{}/{}]".format(np.mean(precision_5), np.mean(recall_5)))
        print("pre@10, rec@10: [{}/{}]".format(np.mean(precision_10), np.mean(recall_10)))
        print("pre@15, rec@15: [{}/{}]".format(np.mean(precision_15), np.mean(recall_15)))
        print("pre@20, rec@20: [{}/{}]".format(np.mean(precision_20), np.mean(recall_20)))

    result.write('\t'.join([str(cnt), str(uid), str(np.mean(precision_5)), str(np.mean(recall_5))]) + '\n')
    result.write('\t'.join([str(cnt), str(uid), str(np.mean(precision_10)), str(np.mean(recall_10))]) + '\n')
    result.write('\t'.join([str(cnt), str(uid), str(np.mean(precision_15)), str(np.mean(recall_15))]) + '\n')
    result.write('\t'.join([str(cnt), str(uid), str(np.mean(precision_20)), str(np.mean(recall_20))]) + '\n')
    print('done：', time.time() - st_time)
